"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports._globalOpts = undefined;

var _typeof = typeof Symbol === "function" && typeof Symbol.iterator === "symbol" ? function (obj) { return typeof obj; } : function (obj) { return obj && typeof Symbol === "function" && obj.constructor === Symbol && obj !== Symbol.prototype ? "symbol" : typeof obj; }; // @flow

exports.configureSentinels = configureSentinels;
exports.default = makeSentinel;
exports.isSentinel = isSentinel;

var _isWeakMapAvailable = require("./isWeakMapAvailable");

var _isWeakMapAvailable2 = _interopRequireDefault(_isWeakMapAvailable);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

// =============================================================================
// Global configuration
// =============================================================================

/*:: type FullSentinelOpts = {|
  shouldIgnore: ShouldIgnore,
  mutationHandler: MutationHandler,
|};*/
/*:: export type SentinelOpts = {|
  shouldIgnore?: ShouldIgnore | void,
  mutationHandler?: MutationHandler | void,
|};*/
/*:: export type ShouldIgnore = mixed => boolean;*/


// Flow core.js uses the weak Object types for these.
/* eslint-disable flowtype/no-weak-types */
/*:: export type MutationHandler = Mutation => void;*/

/* eslint-enable flowtype/no-weak-types */

/*:: export type Mutation =
  | {|
      type: "defineProperty",
      target: Observable,
      property: string,
      descriptor: Object,
    |}
  | {|
      type: "deleteProperty",
      target: Observable,
      property: string,
    |}
  | {|
      type: "set",
      target: Observable,
      property: string,
      value: any,
    |}
  | {|
      type: "setPrototypeOf",
      target: Observable,
      property: "[[Prototype]]",
      prototype: ?Object,
    |};*/


// Visible for testing
/*:: type Observable = {} | (() => mixed);*/
var _globalOpts /*: FullSentinelOpts*/ = exports._globalOpts = {
  shouldIgnore: falseFn,
  mutationHandler: _defaultMutationHandler
};

function configureSentinels(opts /*: SentinelOpts*/) {
  _globalOpts.shouldIgnore = typeof opts.shouldIgnore === "function" ? opts.shouldIgnore : falseFn;
  _globalOpts.mutationHandler = typeof opts.mutationHandler === "function" ? opts.mutationHandler : _defaultMutationHandler;
}

function falseFn() {
  return false;
}

// =============================================================================
// makeSentinel
// =============================================================================

// Map a non-sentinel value to its sentinel counterpart.
// By using this cache we can guarantee the invariant that for a single object
// reference, there will ever only be one corresponding sentinel.
//
// There is no way to type the WeakMap to enforce that a key/value pair has the
// same type for the key and value, but allow different types of key/value
// pairs in the map. Therefore we type the value as `any` and make sure that in
// our code we enforce that we only cache <T, T> key/value pairs (or more
// specifically, <T, Proxy<T>> pairs).
/*:: type SentinelCache = WeakMap<Observable, any>;*/

var _sentinelCache /*: SentinelCache | void*/ = _isWeakMapAvailable2.default ? new WeakMap() : undefined;

// This really just needs to be a set, but not all browsers that support Proxy
// support WeakSet, but they all support WeakMap.
/*:: type KnownSentinels = WeakMap<Observable, boolean>;*/

var _knownSentinels /*: KnownSentinels | void*/ = _isWeakMapAvailable2.default ? new WeakMap() : undefined;

/**
 * Returns a Proxy for the value if it is possible to watch for mutations on
 * the value. Otherwise, return the value itself.
 *
 * Since our sentinel Proxy does not change the object in any way, including
 * the flow types, we can force the return type to always be T.
 */
function makeSentinel /*:: <T>*/(value /*: T*/) /*: T*/ {
  if (typeof Proxy === "undefined" || _sentinelCache == null || _knownSentinels == null || value == null || (typeof value === "undefined" ? "undefined" : _typeof(value)) !== "object" && typeof value !== "function" || _globalOpts.shouldIgnore(value)) {
    return value;
  }

  var knownSentinels = _knownSentinels;
  var sentinelCache = _sentinelCache;

  if (knownSentinels.has(value)) {
    return value;
  }

  var cachedSentinel = sentinelCache.get(value);
  if (cachedSentinel != null) {
    return cachedSentinel;
  }

  var sentinel = new Proxy(value, {
    get: function get(target, property, receiver) {
      var targetVal = target[property];
      if (_canMakeNestedSentinel(target, property, targetVal, knownSentinels)) {
        return makeSentinel(targetVal);
      } else {
        return targetVal;
      }
    },
    defineProperty: function defineProperty(target, property, descriptor) {
      var curDescriptor = Object.getOwnPropertyDescriptor(target, property);
      // We consider any property that has a getter as mutating because that
      // getter can return any value.
      if (curDescriptor == null || curDescriptor.value !== descriptor.value || descriptor.get) {
        _globalOpts.mutationHandler({
          type: "defineProperty",
          target: target,
          property: property,
          descriptor: descriptor
        });
      }
      Object.defineProperty(target, property, descriptor);
      return true;
    },
    deleteProperty: function deleteProperty(target, property) {
      if (Object.prototype.hasOwnProperty.call(target, property)) {
        _globalOpts.mutationHandler({
          type: "deleteProperty",
          target: target,
          property: property
        });
      }
      return delete target[property];
    },
    set: function set(target, property, value, receiver) {
      if (!_valueEq(target[property], value, sentinelCache)) {
        _globalOpts.mutationHandler({
          type: "set",
          target: target,
          property: property,
          value: value
        });
      }
      target[property] = value;
      return true;
    },
    setPrototypeOf: function setPrototypeOf(target, prototype) {
      if (Object.getPrototypeOf(target) !== prototype) {
        _globalOpts.mutationHandler({
          type: "setPrototypeOf",
          target: target,
          property: "[[Prototype]]",
          prototype: prototype
        });
      }
      Object.setPrototypeOf(target, prototype);
      return true;
    }
  });

  sentinelCache.set(value, sentinel);
  knownSentinels.set(sentinel, true);
  return sentinel;
}

/**
 * The `get` handler for a Proxy must enforce certain invariants:
 * https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Proxy/handler/get
 *
 * However, we only need to enforce the first invariant. The second invariant
 * deals with `undefined` values, which we don't try to wrap with a Sentinel
 * anyways.
 *
 * The invariant is enforced only on target's own properties, which means that
 * we don't need to check if a non-writable and non-configurable property with
 * the same name exists up the prototype chain.
 */
function _canMakeNestedSentinel /*:: <T>*/(target /*: T*/, property /*: string*/, targetVal /*: mixed*/, knownSentinels /*: KnownSentinels*/) /*: boolean*/ {
  if (targetVal == null || (typeof targetVal === "undefined" ? "undefined" : _typeof(targetVal)) !== "object" && typeof targetVal !== "function" || knownSentinels.has(targetVal)) {
    return false;
  }
  var descriptor = Object.getOwnPropertyDescriptor(target, property);
  return descriptor && (descriptor.writable || descriptor.configurable);
}

/**
 * Returns true if curValue is equal to newValue or if newValue is the sentinel
 * version of curValue.
 */
function _valueEq(curValue /*: mixed*/, newValue /*: mixed*/, sentinelCache /*: SentinelCache*/) /*: boolean*/ {
  if (curValue === newValue) {
    return true;
  }
  if (newValue == null || (typeof newValue === "undefined" ? "undefined" : _typeof(newValue)) !== "object") {
    // newValue can't be a sentinel
    return false;
  }
  // It is legal to call WeakMap.get with any type, but flow doesn't like
  // it. Just cast to 'any' so we can avoid unnecessary typeof checking.
  return sentinelCache.get((curValue /*: any*/)) === newValue;
}

function _defaultMutationHandler(mutation /*: Mutation*/) /*: void*/ {
  // eslint-disable-next-line no-console
  console.warn("Mutation detected by a sentinel!", mutation);
}

function isSentinel(value /*: mixed*/) /*: boolean*/ {
  return value != null && ((typeof value === "undefined" ? "undefined" : _typeof(value)) === "object" || typeof value === "function") && _knownSentinels != null && _knownSentinels.has(value);
}